"""
Manifestly is a Python library for creating and managing manifest files.
"""
from .core import Manifest

__version__ = "0.2.4"


def get_version():
    """
    Get the version of the library
    :return: MAJOR.MINOR.PATCH version
    """
    return __version__


__all__ = ["Manifest", "get_version"]



"""
Manifestly CLI

Usage:
    manifestly generate <directory> [--hash-algorithm=<hash_algorithm>] [--output-file=<output_file>]
    manifestly sync <source_manifest> <target_manifest> <source_directory> <target_directory>
    manifestly compare <manifest1> <manifest2>
    manifestly patch <source_manifest> <target_manifest> <output_patch_file>
    manifestly pzip <source_manifest> <target_manifest> <output_zip_file>
"""
import click
import fsspec

from manifestly import settings
from .core import Manifest


@click.group()
def cli():
    """
    Manifestly CLI
    :return:
    """
    pass


@cli.command('generate')
@click.argument('directory')
@click.option('--hash-algorithm', default=settings.DEFAULT_HASH_ALGORITHM)
@click.option('--output-file', default=None)
def generate_cmd(directory, hash_algorithm, output_file):
    """
    Generate a manifest file
    :param directory: The directory to generate the manifest for
    :param hash_algorithm: The hash algorithm to use
    :param output_file: The output file
    """
    if output_file is None:
        output_file = Manifest.default_manifest_file(directory)
    else:
        output_file = fsspec.open(output_file, 'w')
    Manifest.generate(directory, manifest_file=output_file, hash_algorithm=hash_algorithm)
    click.echo(f"Manifest saved to {output_file.path}")


@cli.command('changed')
@click.argument('manifest')
@click.option('--root', default=None)
def changed_cmd(manifest, root):
    """
    Print the files that have changed
    :param manifest: The manifest file
    :param root: The root directory
    """
    m = Manifest(manifest, root=root)
    _changed = m.changed()
    if not any(_changed.values()):
        click.echo('No files have changed')
        return
    click.echo('Changed files:')
    for category, files in _changed.items():
        if not files:
            continue
        click.echo(f"{category}:")
        for file in files:
            click.echo(f"  {file}")


@cli.command('refresh')
@click.argument('manifest')
@click.option('--root', default=None)
def refresh_cmd(manifest, root=None):
    """
    Refresh the manifest file
    :param manifest: The manifest file
    :param root: The root directory
    """
    m = Manifest(manifest, root=root)
    m.refresh()
    click.echo("Manifest refreshed")


@cli.command('sync')
@click.argument('source_manifest')
@click.argument('target_manifest')
@click.option('--refresh', is_flag=True, help="Refresh the source manifest first.", default=False)
@click.option('--dry-run', is_flag=True, help="Perform a dry run", default=False)
@click.option('--source_directory', default=None, help="The source directory")
@click.option('--target_directory', default=None, help="The target directory")
def sync_cmd(source_manifest, target_manifest, source_directory=None, target_directory=None, refresh=False,
             dry_run=False):
    """
    Sync two directories using manifest files
    :param source_manifest: Source manifest file
    :param target_manifest: Target manifest file
    :param source_directory: The source directory
    :param target_directory: The target directory
    :param refresh: Refresh the source manifest first
    :param dry_run: Perform a dry run
    """
    s_manifest = Manifest(source_manifest, root=source_directory)
    if refresh:
        s_manifest.refresh()
    t_manifest = Manifest(target_manifest, root=target_directory)
    s_manifest.sync(t_manifest, dry_run=dry_run)
    if dry_run:
        click.echo(f"Dry run completed for {source_directory} to {target_directory}")
    else:
        click.echo(f"Synced {s_manifest.root} with {t_manifest.root}")


@cli.command('compare')
@click.argument('source_manifest')
@click.argument('target_manifest')
def compare_cmd(source_manifest, target_manifest):
    """
    Compare two manifest files and print the diff
    :param source_manifest: The source manifest file
    :param target_manifest: The target manifest file
    """
    s_manifest = Manifest(source_manifest)
    diff = s_manifest.diff(target_manifest)
    click.echo(diff)


@cli.command('patch')
@click.argument('source_manifest')
@click.argument('target_manifest')
@click.argument('output_patch_file')
def patch_cmd(source_manifest, target_manifest, output_patch_file):
    """
    Generate a patch file
    :param source_manifest: The source manifest file
    :param target_manifest:  The target manifest file
    :param output_patch_file: The output patch file
    :return:
    """
    s_manifest = Manifest(source_manifest)
    s_manifest.patch(target_manifest, output_patch_file)
    click.echo(f"Patch saved to {output_patch_file}")


@cli.command('pzip')
@click.argument('source_manifest')
@click.argument('target_manifest')
@click.argument('output_zip_file')
def pzip_cmd(source_manifest, target_manifest, output_zip_file):
    """
    Generate a zip file with the differences
    :param source_manifest: The source manifest file
    :param target_manifest: The target manifest file
    :param output_zip_file: Path to the output zip file
    """
    s_manifest = Manifest(source_manifest)
    s_manifest.pzip(target_manifest, output_zip_file)
    click.echo(f"Zip file saved to {output_zip_file}")


if __name__ == '__main__':
    cli()  # pragma: no cover



import json
import pathlib
import shutil
import tempfile
import unittest

import fsspec

from manifestly import Manifest, settings
from manifestly.core import ManifestlyIgnore

NO_CHANGES = {'added': {}, 'removed': {}, 'changed': {}}


class ManifestlyGeneralTestCase(unittest.TestCase):

    def test_version(self):
        import manifestly
        self.assertEqual(manifestly.get_version(), manifestly.__version__)


class ManifestlyManifestTestCase(unittest.TestCase):

    def setUp(self):
        """
        Common setup for all tests
        """

        self._tmpdir = pathlib.Path(tempfile.mkdtemp())
        self._syncdir = pathlib.Path(tempfile.mkdtemp())

        # This has the test files, do not create manifests in this as it is checked into git
        # Copy files into the tmpdir
        self.manifest_dir = pathlib.Path(__file__).parent / 'test_files'

    def copy_test_files(self, dest):
        """
        Copy over files from the source to the destination
        """
        # Make sure the destination exists
        dest = pathlib.Path(dest)
        dest.mkdir(parents=True, exist_ok=True)
        for f in self.manifest_dir.iterdir():
            if f.is_dir():
                shutil.copytree(str(f), str(dest / f.name))
            else:
                shutil.copy(str(f), str(dest))

    def tearDown(self):
        """
        Common teardown for all tests
        """
        shutil.rmtree(str(self._tmpdir))
        shutil.rmtree(str(self._syncdir))

    def test_manifest_creation(self):
        manifest_dir = pathlib.Path(__file__).parent / 'test_files'
        _manifest_file = self._tmpdir / '.manifest.json'
        m = Manifest(str(_manifest_file), root=str(manifest_dir))
        m.refresh()
        self.assertTrue(_manifest_file.exists())
        self.assertTrue(m.manifest)
        print(m.manifest)

    def test_manifest_sync(self):
        _sync_manifest = self._syncdir / '.manifestly.json'

        _manifest_file = self._tmpdir / '.manifest.json'
        m = Manifest(str(_manifest_file), root=str(self.manifest_dir))
        m.refresh()
        self.assertTrue(_manifest_file.exists())

        self.assertFalse(_sync_manifest.exists())
        m.sync(Manifest(str(_sync_manifest), root=str(self._syncdir)), dry_run=True)
        self.assertTrue(_sync_manifest.exists())
        self.assertEqual(_sync_manifest.read_text(), '{}')
        m.sync(Manifest(str(_sync_manifest), root=str(self._syncdir)))
        self.assertTrue(_sync_manifest.exists())
        print(m.manifest)

        # Load from directory
        m2 = Manifest(str(self._syncdir))
        self.assertTrue(m2.manifest)

    def test_resolve_root(self):
        _sync_manifest = self._syncdir / '.manifestly.json'

        _manifest_file = self._tmpdir / '.manifest.json'
        self.copy_test_files(self._tmpdir)
        self.copy_test_files(self._syncdir)

        m = Manifest(str(_manifest_file), root=str(self.manifest_dir))
        m.sync(Manifest(str(_sync_manifest), root=str(self._syncdir)))
        self.assertTrue(_sync_manifest.exists())

        # Resolve root from valid manifest file
        m2 = Manifest(str(_manifest_file))
        m2.refresh()
        self.assertTrue(m2.manifest)
        self.assertTrue(m2.root, str(self._tmpdir))

    def test_generation(self):
        _copy_dir = self._tmpdir / 'test_files'
        self.copy_test_files(_copy_dir)

        m = Manifest.generate(str(_copy_dir), str(_copy_dir / settings.MANIFEST_NAME))
        m.refresh()
        self.assertTrue(m.manifest)

        # Generate based of the original files
        m2 = Manifest(str(self._tmpdir / '.manifestly.json'), root=str(self.manifest_dir))
        m2.refresh()
        self.assertTrue(m2.manifest)

        self.assertEqual(m2.changed(), NO_CHANGES)

        # Test the generated manifest
        self.assertEqual(m.manifest, m2.manifest)

        # Test the diff
        diff = m.diff(m2)
        self.assertEqual(diff, NO_CHANGES)

        self.assertEqual(m, m2)

        self.assertTrue('.manifestlyignore' in m)
        self.assertNotEqual(m, 'test')
        self.assertTrue('.manifestlyignore' in m.keys())
        self.assertTrue('0cc6c7041e35947e9cb27e32f237ed4db36745ea362000ad4d377e2653a63775' in m.values())

    def test_changes(self):
        _orig_dir = self._tmpdir / 'orig_files'
        _change_dir = self._tmpdir / 'change_files'
        self.copy_test_files(_orig_dir)
        self.copy_test_files(_change_dir)

        orig_manifest = Manifest.generate(str(_orig_dir), str(_orig_dir / settings.MANIFEST_NAME))
        self.assertTrue(orig_manifest.manifest)
        orig_manifest.save()
        _manifest_file = orig_manifest._reopen()
        self.assertEqual(orig_manifest.manifest, json.loads(_manifest_file.open().read()))
        _old_manifest = orig_manifest.manifest
        orig_manifest.refresh()
        self.assertEqual(orig_manifest.manifest, _old_manifest)

        change_manifest = Manifest.generate(str(_change_dir), str(_change_dir / settings.MANIFEST_NAME))
        original_change_manifest = change_manifest.manifest.copy()
        change_manifest.root = None
        change_manifest.refresh()
        self.assertEqual(change_manifest.manifest, original_change_manifest)
        change_manifest.save()

        # No changes
        self.assertEqual(orig_manifest.changed(), NO_CHANGES)
        self.assertEqual(change_manifest.changed(), NO_CHANGES)

        # Remove the test_css.css file
        (_change_dir / 'test_css.css').unlink()
        # Add a file
        _f = _change_dir / 'new_file.txt'
        _f.write_text('test')

        # Update a file
        _uf = _change_dir / 'subdirectory' / 'sub_ts.ts'
        _uf.write_text(_uf.read_text() + '\n// New line')
        changes = change_manifest.changed()
        self.assertNotEqual(changes, NO_CHANGES)
        for key in NO_CHANGES.keys():
            # All categories should have changes
            self.assertTrue(changes[key])

        self.assertEqual(orig_manifest.changed(), NO_CHANGES)

        # You have to refresh/regenerate the manifest to get the changes
        change_manifest.refresh()

        # Test the diff
        # Should be able to pass in the manifest object or the path to the manifest file/directory
        diff = change_manifest.diff(str(_orig_dir))
        diff2 = change_manifest.diff(orig_manifest)
        self.assertEqual(diff, diff2)

        self.assertNotEqual(diff, NO_CHANGES)
        for key in NO_CHANGES.keys():
            # All categories should have changes
            self.assertTrue(diff[key])

        # Create the patch
        patch_file = self._tmpdir / 'patch.json'
        patch = change_manifest.patch(str(_orig_dir), str(patch_file))
        self.assertTrue(patch)
        self.assertTrue(patch_file.exists())

        # Create a pzip (patch zip) with changed files, then check the patch
        zip_file = self._tmpdir / 'patch.zip'
        # Pass in the directory so we hit the case where we have a directory
        # You usually pass in the manifest object
        change_manifest.pzip(str(_orig_dir), str(zip_file))
        self.assertTrue(zip_file.exists())
        self.assertTrue(zip_file.stat().st_size > 0)
        import zipfile
        extracted_dir = self._tmpdir / 'extracted'
        with zipfile.ZipFile(str(zip_file), 'r') as z:
            z.extractall(str(extracted_dir))
        self.assertTrue((self._tmpdir / 'extracted').exists())

        # Check the files in the extracted zip
        self.assertTrue((extracted_dir / 'new_file.txt').exists())
        self.assertTrue((extracted_dir / 'subdirectory' / 'sub_ts.ts').exists())
        self.assertFalse((extracted_dir / 'test_css.css').exists())

        # Synchronize the changes to the original directory
        change_manifest.sync(str(_orig_dir), dry_run=True)

        # Remove the added file and make sure it doesn't throw errors when we sync
        (_change_dir / 'new_file.txt').unlink()
        self.assertFalse((_change_dir / 'new_file.txt').exists())
        change_manifest.sync(orig_manifest)
        self.assertFalse((_orig_dir / 'new_file.txt').exists())

        root = fsspec.open(str(change_manifest.root))
        new_manifest = Manifest.generate(str(_change_dir),
                                         str(_change_dir / settings.MANIFEST_NAME), root_path=root)
        new_manifest.save()

        # Test the diff
        diff = new_manifest.diff(orig_manifest)
        self.assertEqual(diff, NO_CHANGES)

    def test_ignore(self):
        _copy_dir = self._tmpdir / 'test_files'
        self.copy_test_files(_copy_dir)

        m = Manifest.generate(str(_copy_dir), str(_copy_dir / settings.MANIFEST_NAME))
        self.assertFalse(settings.MANIFEST_NAME in m.manifest)
        m.refresh()
        self.assertFalse(settings.MANIFEST_NAME in m.manifest)
        self.assertTrue(m.manifest)

        # Make sure the ignored files are not present in the manifest
        manifest = m.manifest
        self.assertFalse(settings.MANIFEST_NAME in manifest)
        self.assertFalse('subdirectory/ignore_me.md' in manifest.keys())
        self.assertFalse('subdirectory/dir_ignore/also_should_be_ignored.py' in manifest.keys())

        # Remove the ignore file and then refresh
        _ignore_file = m.ignore.ignore_file
        # This is an OpenFile, so we need to get the path
        source, path = fsspec.core.url_to_fs(_ignore_file.path)
        source.rm(path)
        self.assertFalse(source.exists(path))

        m2 = Manifest(str(_copy_dir))
        m2.refresh()
        self.assertTrue(m2.manifest)

        # Still should not have the manifest name in the manifest, even without an ignore file
        self.assertFalse(settings.MANIFEST_NAME in m2.manifest)
        # These should now be in the manifest
        self.assertTrue('subdirectory/ignore_me.md' in m2)
        self.assertTrue('subdirectory/dir_ignore/also_should_be_ignored.py' in m2)

    def test_bad_manifest(self):
        _orig_dir = self._tmpdir / 'orig_files'
        self.copy_test_files(_orig_dir)

        m1 = Manifest.generate(str(_orig_dir), str(_orig_dir / settings.MANIFEST_NAME))
        m1.save()
        _manifest_file = m1._reopen()
        self.assertEqual(m1.manifest, json.loads(_manifest_file.open().read()))

        # Corrupt the manifest
        m1._reopen('w').open().write('bad json')
        m1.load()
        self.assertEqual(m1.manifest, {})

        m1._reopen('w').open().write('')
        m1.load()
        self.assertEqual(m1.manifest, {})


class TestManifestlyIgnore(unittest.TestCase):

    def setUp(self):
        self._tmpdir = pathlib.Path(tempfile.mkdtemp())
        self.manifest_dir = pathlib.Path(__file__).parent / 'test_files'

    def tearDown(self):
        shutil.rmtree(str(self._tmpdir))

    def test_manifestly_ignore(self):
        ignore_file = Manifest.default_ignore_file(str(self.manifest_dir))
        # Test if the OpenFile exists
        source, path = fsspec.core.url_to_fs(ignore_file.path)
        self.assertTrue(source.exists(path))

        # Create the Manifestly Ignore object
        _ignore = ManifestlyIgnore(ignore_file)
        self.assertTrue(_ignore)
        self.assertTrue(_ignore.ignore_patterns)
        self.assertIn('.DS_Store', _ignore.ignore_patterns)

    def test_manifestly_ignore_dne(self):
        tmp_dir = self._tmpdir
        ignore_file = Manifest.default_ignore_file(str(tmp_dir))
        # Create the Manifestly Ignore object
        _ignore = ManifestlyIgnore(ignore_file.path)
        self.assertTrue(_ignore)
        self.assertTrue(_ignore.ignore_patterns)
        self.assertNotIn('.DS_Store', _ignore.ignore_patterns)

        _full_path = tmp_dir / '.DS_Store'
        _ignore.add_ignore_pattern(fsspec.open(str(_full_path)))
        self.assertIn('.DS_Store', _ignore.ignore_patterns)

        # Add a directory to ignore
        _ignore.add_ignore_pattern('test_dir/')
        self.assertIn('test_dir/', _ignore.ignore_patterns)
        self.assertTrue(_ignore.should_ignore('test_dir/test_file.txt'))
        self.assertTrue(_ignore.should_ignore('test_dir/', ))


if __name__ == '__main__':
    unittest.main()  # pragma: no cover



"""
Manifestly resolved settings from environment variables
"""
import os

DEFAULT_HASH_ALGORITHM = os.getenv('MANIFESTLY_HASH_ALGORITHM', 'sha256')
MANIFEST_NAME = os.getenv('MANIFESTLY_NAME', '.manifestly.json')
MANIFESTLY_IGNORE = os.getenv('MANIFESTLY_IGNORE', '.manifestlyignore')
CHUNK_SIZE = int(os.getenv('MANIFESTLY_CHUNK_SIZE', 8192))



import pathlib
import shutil
import tempfile
import unittest

from click.testing import CliRunner

from manifestly import settings
from manifestly.cli import cli


class ManifestlyCLITestCase(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()
        self._tmpdir = pathlib.Path(tempfile.mkdtemp())
        self.manifest_dir = pathlib.Path(__file__).parent / 'test_files'
        self.copy_test_files(self._tmpdir)

    def tearDown(self):
        shutil.rmtree(str(self._tmpdir))

    def copy_test_files(self, dest):
        """
        Copy over files from the source to the destination
        """
        for f in self.manifest_dir.iterdir():
            if f.is_dir():
                shutil.copytree(str(f), str(dest / f.name))
            else:
                shutil.copy(str(f), str(dest))

    def test_generate_cmd(self):
        result = self.runner.invoke(cli, ['generate', str(self._tmpdir)])
        self.assertEqual(result.exit_code, 0)
        manifest_file = self._tmpdir / settings.MANIFEST_NAME
        self.assertTrue(manifest_file.exists())

    def test_generate_cmd2(self):
        manifest_file = str(self._tmpdir / settings.MANIFEST_NAME)
        result = self.runner.invoke(cli, ['generate', str(self._tmpdir), '--output-file', manifest_file])
        self.assertEqual(result.exit_code, 0)
        manifest_file = self._tmpdir / settings.MANIFEST_NAME
        self.assertTrue(manifest_file.exists())

    def test_changed_cmd(self):
        # First, generate a manifest
        self.runner.invoke(cli, ['generate', str(self._tmpdir)])

        # Now, change a file
        (self._tmpdir / 'new_file.txt').write_text('new content')

        result = self.runner.invoke(cli, ['changed', str(self._tmpdir / settings.MANIFEST_NAME)])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Changed files:', result.output)

    def test_changed_cmd_no_changes(self):
        # First, generate a manifest
        self.runner.invoke(cli, ['generate', str(self._tmpdir)])

        result = self.runner.invoke(cli, ['changed', str(self._tmpdir / settings.MANIFEST_NAME)])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('No files have changed', result.output)
        self.assertNotIn('Changed files:', result.output)

    def test_refresh_cmd(self):
        self.runner.invoke(cli, ['generate', str(self._tmpdir)])
        manifest_file = self._tmpdir / settings.MANIFEST_NAME
        self.assertTrue(manifest_file.exists())

        result = self.runner.invoke(cli, ['refresh', str(manifest_file)])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Manifest refreshed', result.output)

    def test_sync_cmd(self):
        self.runner.invoke(cli, ['generate', str(self._tmpdir)])
        manifest_file = self._tmpdir / settings.MANIFEST_NAME

        sync_dir = pathlib.Path(tempfile.mkdtemp())
        sync_manifest_file = sync_dir / 'sync_manifest.json'

        result = self.runner.invoke(cli, [
            'sync', str(manifest_file), str(sync_manifest_file),
            '--source_directory', str(self._tmpdir), '--target_directory', str(sync_dir), '--dry-run'
        ])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Dry run completed', result.output)
        shutil.rmtree(str(sync_dir))

    def test_sync_cmd_refresh(self):
        self.runner.invoke(cli, ['generate', str(self._tmpdir)])
        manifest_file = self._tmpdir / settings.MANIFEST_NAME

        sync_dir = pathlib.Path(tempfile.mkdtemp())
        sync_manifest_file = sync_dir / 'sync_manifest.json'

        result = self.runner.invoke(cli, [
            'sync', str(manifest_file), str(sync_manifest_file),
            '--source_directory', str(self._tmpdir), '--target_directory', str(sync_dir), '--refresh'])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Synced', result.output)
        shutil.rmtree(str(sync_dir))

    def test_compare_cmd(self):
        self.runner.invoke(cli, ['generate', str(self._tmpdir)])
        manifest_file = self._tmpdir / settings.MANIFEST_NAME

        compare_dir = pathlib.Path(tempfile.mkdtemp())
        self.copy_test_files(compare_dir)
        compare_manifest_file = compare_dir / 'compare_manifest.json'
        self.runner.invoke(cli, ['generate', str(compare_dir)])

        result = self.runner.invoke(cli, ['compare', str(manifest_file), str(compare_manifest_file)])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('added', result.output)
        shutil.rmtree(str(compare_dir))

    def test_patch_cmd(self):
        self.runner.invoke(cli, ['generate', str(self._tmpdir)])
        manifest_file = self._tmpdir / settings.MANIFEST_NAME

        patch_dir = pathlib.Path(tempfile.mkdtemp())
        self.copy_test_files(patch_dir)
        patch_manifest_file = patch_dir / 'patch_manifest.json'
        self.runner.invoke(cli, ['generate', str(patch_dir)])

        output_patch_file = self._tmpdir / 'output_patch.json'

        result = self.runner.invoke(cli, [
            'patch', str(manifest_file), str(patch_manifest_file), str(output_patch_file)
        ])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Patch saved', result.output)
        self.assertTrue(output_patch_file.exists())
        shutil.rmtree(str(patch_dir))

    def test_pzip_cmd(self):
        self.runner.invoke(cli, ['generate', str(self._tmpdir)])
        manifest_file = self._tmpdir / settings.MANIFEST_NAME

        pzip_dir = pathlib.Path(tempfile.mkdtemp())
        self.copy_test_files(pzip_dir)
        pzip_manifest_file = pzip_dir / 'pzip_manifest.json'
        self.runner.invoke(cli, ['generate', str(pzip_dir)])

        output_zip_file = self._tmpdir / 'output.zip'

        result = self.runner.invoke(cli, [
            'pzip', str(manifest_file), str(pzip_manifest_file), str(output_zip_file)
        ])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Zip file saved', result.output)
        self.assertTrue(output_zip_file.exists())
        shutil.rmtree(str(pzip_dir))


if __name__ == '__main__':
    unittest.main()  # pragma: no cover



"""
Manifestly is a Python library for generating, loading, comparing, and syncing file manifests.

This is the core module that provides the core functionality for generating, loading, comparing,
and syncing file manifests.
"""

import fnmatch
import hashlib
import json
from itertools import chain
from json import JSONDecodeError
from typing import Union

import fsspec
from fsspec.core import OpenFile

from manifestly import settings


class ManifestlyIgnore:
    """
    Handles a .manifestlyignore file that works like a .gitignore file
    You can specify files or directories to ignore.
    This loads the .manifestlyignore file and provides a method to check if a file should be ignored.
    """

    def __init__(self, ignore_file: Union[str, OpenFile]):
        if isinstance(ignore_file, str):
            ignore_file = fsspec.open(ignore_file, 'r')
        self.ignore_file = ignore_file
        self.ignore_patterns = self.load_ignore_patterns()

    def load_ignore_patterns(self):
        """
        Load the ignore patterns from the .manifestlyignore file
        """
        ignore = [settings.MANIFEST_NAME]
        try:
            with self.ignore_file.open() as f:
                return ignore + [self.normalize_path(p) for p in f.read().splitlines()]
        except FileNotFoundError:
            return ignore

    def should_ignore(self, file_path: str) -> bool:
        """
        Check if a file should be ignored.
        :param file_path: The path to the file.
        :return: True if the file should be ignored, False otherwise.
        """
        # Normalize the file path to always use forward slashes
        normalized_path = self.normalize_path(file_path)

        for pattern in self.ignore_patterns:
            _pattern = pattern.strip('/')

            if any(fnmatch.fnmatch(part, _pattern) for part in normalized_path.split('/')):
                return True

        return False

    @staticmethod
    def normalize_path(path: str) -> str:
        """
        Normalize a path to always use forward slashes
        :param path: The path to normalize
        :return: The normalized path
        """
        return path.replace('\\', '/')

    def add_ignore_pattern(self, name: Union[str, OpenFile]):
        """
        Add an ignore pattern
        :param name: pattern to ignore
        """
        if isinstance(name, OpenFile):
            name = self.normalize_path(name.path).split('/')[-1]
        if name not in self.ignore_patterns:
            self.ignore_patterns.append(self.normalize_path(name))


class Manifest:
    """
    A class to represent a manifest
    """

    def __init__(self, manifest_file: Union[str, OpenFile], manifest: dict = None,
                 root: str = None, ignore: ManifestlyIgnore = None):
        if isinstance(manifest_file, str):
            manifest_file = fsspec.open(manifest_file)
        self.manifest_file: OpenFile = manifest_file
        self.manifest: dict = manifest
        self.root = root
        if self.manifest is None:
            self.load()
        if ignore is None:
            ignore = ManifestlyIgnore(self.default_ignore_file(self.root))
        self.ignore = ignore

    def _reopen(self, mode='r'):
        return fsspec.open(self.manifest_file.path, mode)

    def save(self):
        """
        Save the manifest to a file
        :param file_path: The path to the manifest file
        """
        _file = self._reopen('w')
        # Make sure directories exist
        _file.fs.makedirs(_file.fs._parent(_file.path), exist_ok=True)

        with _file.open() as f:
            f.seek(0)
            json.dump(self.manifest, f, indent=2)

    def items(self):
        """
        Get the items in the manifest
        :return:
        """
        return self.manifest.items()

    def keys(self):
        """
        Get the keys in the manifest
        :return:
        """
        return self.manifest.keys()

    def values(self):
        """
        Get the values in the manifest
        :return:
        """
        return self.manifest.values()

    def __eq__(self, other):
        """
        Check if the manifest is equal to another manifest
        :param other: The other manifest
        :return: True if the manifests are equal, False otherwise
        """
        if not isinstance(other, Manifest):
            return False
        return self.manifest == other.manifest

    def __contains__(self, item):
        """
        Check if the manifest contains a file
        :param item: The file to check
        :return: True if the file is in the manifest, False otherwise
        """
        return item in self.manifest

    def load(self):
        """
        Load a manifest from a file
        :return: The loaded manifest
        """
        try:
            with self._reopen() as f:
                _data = f.read()
                if not _data:
                    self.manifest = {}
                else:
                    self.manifest = json.loads(_data)
        except FileNotFoundError:
            self.manifest = {}
            self.save()
        except IsADirectoryError:
            # The manifest file is a directory, so we need to append the default manifest file name
            self.root = self.manifest_file
            self.manifest_file = self.default_manifest_file(self.manifest_file)
            self.load()
        except JSONDecodeError:
            self.manifest = {}
        if self.root is None:
            # Resolve the root path from the manifest file
            fs, path = fsspec.core.url_to_fs(self.manifest_file.path)
            self.root = fs._parent(path)

    @classmethod
    def default_manifest_file(cls, directory: Union[str, OpenFile]) -> OpenFile:
        """
        Get the default manifest file for a directory.
        The directory must be a string or an fsspec.core.OpenFile object.

        :param directory: The directory
        :return: The path to the manifest file
        """
        fs, path = fsspec.core.url_to_fs(directory)
        manifest_path = fs.sep.join([path.rstrip(fs.sep), settings.MANIFEST_NAME])
        return fsspec.open(manifest_path)

    @classmethod
    def default_ignore_file(cls, directory: Union[str, OpenFile]) -> OpenFile:
        """
        Get the default manifest file for a directory.
        The directory must be a string or an fsspec.core.OpenFile object.

        :param directory: The directory
        :return: The path to the manifest file
        """
        fs, path = fsspec.core.url_to_fs(directory)
        manifest_path = fs.sep.join([path.rstrip(fs.sep), settings.MANIFESTLY_IGNORE])
        return fsspec.open(manifest_path, 'r')

    @classmethod
    def generate(cls, directory, manifest_file: Union[str, OpenFile] = None, root_path: str = None,
                 hash_algorithm=settings.DEFAULT_HASH_ALGORITHM, ignore: ManifestlyIgnore = None) -> 'Manifest':
        """
        Generate a manifest for a directory
        :param directory: The directory to generate the manifest for
        :param manifest_file: Optional path to the manifest file
        :param root_path: The root path to use (all paths will be relative to this)
        :param hash_algorithm: Hash algorithm to use
        :param ignore: The ignore file
        :return: The generated manifest
        """
        manifest = {}
        fs, path = fsspec.core.url_to_fs(directory)
        if root_path is None:
            root_path = path
        elif isinstance(root_path, OpenFile):
            root_path = root_path.path

        # Load the ignore file if it exists
        if ignore is None:
            ignore = ManifestlyIgnore(cls.default_ignore_file(directory))
        ignore.add_ignore_pattern(settings.MANIFEST_NAME)
        if manifest_file:
            if isinstance(manifest_file, OpenFile):
                _ignore_file = manifest_file.path
            else:
                _ignore_file = manifest_file
            _ignore_file = _ignore_file.split('/')[-1].split('\\')[-1]
            ignore.add_ignore_pattern(_ignore_file)

        for file_path in fs.find(path):
            if fs.isfile(file_path):
                # Check ignore patterns
                if ignore.should_ignore(file_path):
                    continue
                relative_path = file_path[len(root_path):].lstrip('/')
                manifest[relative_path] = cls.calculate_hash(fs.open(file_path), algorithm=hash_algorithm)

        if manifest_file:
            if not isinstance(manifest_file, OpenFile):
                manifest_file = fsspec.open(manifest_file, 'w')
            elif 'w' not in manifest_file.mode:
                manifest_file = fsspec.open(manifest_file.path, 'w')
            with manifest_file as f:
                json.dump(manifest, f, indent=2)
        return cls(manifest_file, manifest, root=root_path, ignore=ignore)

    def changed(self) -> dict:
        """
        Get the files that have changed
        This returns a dictionary of added, removed, and changed files
        :return: Dictionary of changed files
        """
        self.load()
        fs, path = fsspec.core.url_to_fs(self.root)
        changed = {
            'added': {},
            'removed': {},
            'changed': {}
        }
        for file, _hash in self.manifest.items():
            file_path = f'{path}/{file}'
            if not fs.exists(file_path):
                changed['removed'][file] = _hash
                continue
            _new_hash = self.calculate_hash(fs.open(file_path))
            if _hash != _new_hash:
                changed['changed'][file] = _new_hash
        for file in fs.find(path):
            if self.ignore.should_ignore(file):
                continue
            relative_path = file[len(path):].lstrip('/')
            if relative_path not in self.manifest:
                changed['added'][relative_path] = self.calculate_hash(fs.open(file))
        return changed

    def sync(self, target_manifest, dry_run=False) -> 'Manifest':
        """
        Sync the target directory to match the source manifest
        :param target_manifest: The path to the target manifest file or a Manifest object
        :param dry_run: Perform a dry run
        """
        if isinstance(target_manifest, str):
            target_manifest = Manifest(target_manifest)
        diff = self.diff(target_manifest)

        fs_source, source_path = fsspec.core.url_to_fs(self.root)
        fs_target, target_path = fsspec.core.url_to_fs(target_manifest.root)

        for file in chain(diff['added'].keys(), diff['changed'].keys()):
            source_file = f'{source_path}/{file}'
            target_file = f'{target_path}/{file}'
            fs_target.mkdirs(fs_target._parent(target_file), exist_ok=True)
            # Check if the file still exists before copying
            if not fs_source.exists(source_file):
                print(f'File {source_file} does not exist')
                continue

            if dry_run:
                print(f'Copy {source_file} to {target_file}')
                continue
            with fs_source.open(source_file, 'rb') as src, fs_target.open(target_file, 'wb') as tgt:
                tgt.write(src.read())

        for file in diff['removed']:
            target_file = f'{target_path}/{file}'
            if fs_target.exists(target_file):
                if dry_run:
                    print(f'Remove {target_file}')
                    continue
                fs_target.rm(target_file)

        if not dry_run:
            # Regenerate the target manifest
            target_manifest.refresh()
        return target_manifest

    def refresh(self):
        """
        Regenerate the manifest
        """
        self.manifest = {}
        directory = self.root
        if directory is None:
            # Resolve the directory from the manifest file
            fs, path = fsspec.core.url_to_fs(self.manifest_file)  # noqa
            directory = fs._parent(path)  # noqa
        _m = self.generate(directory=directory, manifest_file=self.manifest_file, root_path=self.root,
                           ignore=self.ignore)
        self.manifest = _m.manifest

    def diff(self, target_manifest) -> dict:
        """
        Compare the manifest with another manifest
        :param target_manifest: The path to the target manifest file or a Manifest object
        :return: A dictionary with the differences
        """
        if not isinstance(target_manifest, Manifest):
            target_manifest = Manifest(target_manifest)
        diff = {
            'added': {},
            'removed': {},
            'changed': {}
        }
        for file, _hash in self.items():
            if file not in target_manifest.manifest:
                diff['added'][file] = _hash
            elif target_manifest.manifest[file] != _hash:
                diff['changed'][file] = _hash
        for file, _hash in target_manifest.items():
            if file not in self.manifest:
                diff['removed'][file] = _hash
        return diff

    def patch(self, target_manifest, output_patch_file) -> dict:
        """
        Generate a patch file that can be used to update the target manifest to match the source manifest.
        :param target_manifest: The path to the target manifest file or a Manifest object
        :param output_patch_file: The path to the output patch/json file
        :return: The diff dictionary
        """
        if not isinstance(target_manifest, Manifest):
            target_manifest = Manifest(target_manifest)
        diff = self.diff(target_manifest)
        with fsspec.open(output_patch_file, 'w') as f:
            json.dump(diff, f, indent=2)
        return diff

    def pzip(self, target_manifest, output_zip_file):
        """
        Generate a zip file containing the files that need to be added or updated to make the target manifest match
        :param target_manifest: The path to the target manifest file or a Manifest object
        :param output_zip_file: The path to the output zip file
        """
        import zipfile
        if not isinstance(target_manifest, Manifest):
            target_manifest = Manifest(target_manifest)
        diff = self.diff(target_manifest)
        fs, _ = fsspec.core.url_to_fs('')
        with fsspec.open(output_zip_file, 'wb') as zf:
            with zipfile.ZipFile(zf, 'w') as zipf:
                for file in chain(diff['added'].keys(), diff['changed'].keys()):
                    # resolve the full path from the root
                    _fpath = f'{self.root}/{file}'
                    with fs.open(_fpath, 'rb') as f:
                        zipf.writestr(file, f.read())
                # Create a '.manifestly.diff' of the diff contents
                with fsspec.open('.manifestly.diff', 'w') as f:
                    f.write(json.dumps(diff))

    @staticmethod
    def calculate_hash(file: OpenFile, algorithm=settings.DEFAULT_HASH_ALGORITHM):
        """
        Calculate the hash of a file
        :param file: The OpenFile object
        :param algorithm: The hash algorithm to use
        :return: The hash of the file
        """
        hasher = hashlib.new(algorithm)
        with file as f:
            while chunk := f.read(settings.CHUNK_SIZE):
                hasher.update(chunk)
        return hasher.hexdigest()
